@@ -270,10 +270,12 @@ def write(stream, msg):
 
         lengths = ([struct.pack('Q', len(frames))] +
                    [struct.pack('Q', len(frame)) for frame in frames])
-        yield stream.write(b''.join(lengths))
+        stream.write(b''.join(lengths))
 
-        for frame in frames:
-            yield stream.write(frame)
+        for frame in frames[:-1]:
+            stream.write(frame)
+
+        yield stream.write(frames[-1])
 
 
 def pingpong(stream):
@@ -317,7 +319,7 @@ def send_recv(stream=None, arg=None, ip=None, port=None, addr=None, reply=True,
         assert not ip and not port
         if PY3 and isinstance(addr, bytes):
             addr = addr.decode()
-        ip, port = addr.split(':')
+        ip, port = addr.rsplit(':', 1)
         port = int(port)
     if PY3 and isinstance(ip, bytes):
         ip = ip.decode()
@@ -379,7 +381,7 @@ class rpc(object):
             if PY3 and isinstance(addr, bytes):
                 addr = addr.decode()
             assert not ip and not port
-            ip, port = addr.split(':')
+            ip, port = addr.rsplit(':', 1)
             port = int(port)
         if PY3 and isinstance(ip, bytes):
             ip = ip.decode()
@@ -449,7 +451,7 @@ def coerce_to_address(o, out=str):
     if PY3 and isinstance(o, bytes):
         o = o.decode()
     if isinstance(o, (unicode, str)):
-        ip, port = o.split(':')
+        ip, port = o.rsplit(':', 1)
         port = int(port)
         o = (ip, port)
     if isinstance(o, list):
