@@ -242,7 +242,8 @@ class Executor(object):
     --------
     distributed.scheduler.Scheduler: Internal scheduler
     """
-    def __init__(self, address, start=True, loop=None, timeout=3):
+    def __init__(self, address, start=True, loop=None, timeout=3,
+                 set_as_default=False):
         self.futures = dict()
         self.refcount = defaultdict(lambda: 0)
         self._should_close_loop = loop is None and start
@@ -250,6 +251,9 @@ class Executor(object):
         self.coroutines = []
         self.id = str(uuid.uuid1())
         self._start_arg = address
+        if set_as_default:
+            self._previous_get = _globals.get('get')
+            dask.set_options(get=self.get)
 
         if start:
             self.start(timeout=timeout)
@@ -426,8 +430,12 @@ class Executor(object):
             sync(self.loop, self.loop.stop)
             self.loop.close()
             self._loop_thread.join(timeout=timeout)
+        with ignoring(AttributeError):
+            dask.set_options(get=self._previous_get)
         if _global_executor[0] is self:
             _global_executor[0] = None
+        if self.get == _globals.get('get'):
+            del _globals['get']
 
     def submit(self, func, *args, **kwargs):
         """ Submit a function application to the scheduler
@@ -567,7 +575,8 @@ class Executor(object):
 
         if (all(map(isqueue, iterables)) or
             all(isinstance(i, Iterator) for i in iterables)):
-            q_out = pyQueue()
+            maxsize = kwargs.pop('maxsize', 0)
+            q_out = pyQueue(maxsize=maxsize)
             t = Thread(target=self._threaded_map, args=(q_out, func, iterables),
                                                   kwargs=kwargs)
             t.daemon = True
@@ -696,7 +705,7 @@ class Executor(object):
             for item in results:
                 qout.put(item)
 
-    def gather(self, futures, errors='raise'):
+    def gather(self, futures, errors='raise', maxsize=0):
         """ Gather futures from distributed memory
 
         Accepts a future, nested container of futures, iterator, or queue.
@@ -725,7 +734,7 @@ class Executor(object):
         Executor.scatter: Send data out to cluster
         """
         if isqueue(futures):
-            qout = pyQueue()
+            qout = pyQueue(maxsize=maxsize)
             t = Thread(target=self._threaded_gather, args=(futures, qout),
                         kwargs={'errors': errors})
             t.daemon = True
@@ -792,7 +801,7 @@ class Executor(object):
             for future in futures:
                 qout.put(future)
 
-    def scatter(self, data, workers=None, broadcast=False):
+    def scatter(self, data, workers=None, broadcast=False, maxsize=0):
         """ Scatter data into distributed memory
 
         Parameters
@@ -805,6 +814,8 @@ class Executor(object):
         broadcast: bool (defaults to False)
             Whether to send each data element to all workers.
             By default we round-robin based on number of cores.
+        maxsize: int (optional)
+            Maximum size of queue if using queues, 0 implies infinite
 
         Returns
         -------
@@ -840,7 +851,7 @@ class Executor(object):
         """
         if isqueue(data) or isinstance(data, Iterator):
             logger.debug("Starting thread for streaming data")
-            qout = pyQueue()
+            qout = pyQueue(maxsize=maxsize)
 
             t = Thread(target=self._threaded_scatter,
                        args=(data, qout),
@@ -941,7 +952,7 @@ class Executor(object):
 
         d = {k: unpack_remotedata(v, byte_keys=True) for k, v in dsk.items()}
         dsk2 = str_graph({k: v[0] for k, v in d.items()})
-        dsk3 = {k: v for k, v in dsk2.items() if (k == v) is not True}
+        dsk3 = {k: v for k, v in dsk2.items() if k is not v}
 
         if restrictions:
             restrictions = keymap(tokey, restrictions)
@@ -1315,6 +1326,7 @@ class Executor(object):
         Examples
         --------
         >>> x, y, z = e.map(inc, [1, 2, 3])  # doctest: +SKIP
+        >>> wait([x, y, z])  # doctest: +SKIP
         >>> e.who_has()  # doctest: +SKIP
         {'inc-1c8dd6be1c21646c71f76c16d09304ea': ['192.168.1.141:46784'],
          'inc-1e297fc27658d7b67b3a758f16bcf47a': ['192.168.1.141:46784'],
@@ -1347,6 +1359,7 @@ class Executor(object):
         Examples
         --------
         >>> x, y, z = e.map(inc, [1, 2, 3])  # doctest: +SKIP
+        >>> wait([x, y, z])  # doctest: +SKIP
         >>> e.has_what()  # doctest: +SKIP
         {'192.168.1.141:46784': ['inc-1c8dd6be1c21646c71f76c16d09304ea',
                                  'inc-fd65c238a7ea60f6a01bf4c8a5fcf44b',
@@ -1364,6 +1377,67 @@ class Executor(object):
             workers = [workers]
         return sync(self.loop, self.scheduler.has_what, keys=workers)
 
+    def stacks(self, workers=None):
+        """ The task queues on each worker
+
+        Parameters
+        ----------
+        workers: list (optional)
+            A list of worker addresses, defaults to all
+
+        Examples
+        --------
+        >>> x, y, z = e.map(inc, [1, 2, 3])  # doctest: +SKIP
+        >>> e.stacks()  # doctest: +SKIP
+        {'192.168.1.141:46784': ['inc-1c8dd6be1c21646c71f76c16d09304ea',
+                                 'inc-fd65c238a7ea60f6a01bf4c8a5fcf44b',
+                                 'inc-1e297fc27658d7b67b3a758f16bcf47a']}
+
+        See Also
+        --------
+        Executor.processing
+        Executor.who_has
+        Executor.has_what
+        Executor.ncores
+        """
+        if (isinstance(workers, tuple)
+            and all(isinstance(i, (str, tuple)) for i in workers)):
+            workers = list(workers)
+        if workers is not None and not isinstance(workers, (list, set)):
+            workers = [workers]
+        return sync(self.loop, self.scheduler.stacks, workers=workers)
+
+    def processing(self, workers=None):
+        """ The tasks currently running on each worker
+
+        Parameters
+        ----------
+        workers: list (optional)
+            A list of worker addresses, defaults to all
+
+        Examples
+        --------
+        >>> x, y, z = e.map(inc, [1, 2, 3])  # doctest: +SKIP
+        >>> e.processing()  # doctest: +SKIP
+        {'192.168.1.141:46784': ['inc-1c8dd6be1c21646c71f76c16d09304ea',
+                                 'inc-fd65c238a7ea60f6a01bf4c8a5fcf44b',
+                                 'inc-1e297fc27658d7b67b3a758f16bcf47a']}
+
+        See Also
+        --------
+        Executor.stacks
+        Executor.who_has
+        Executor.has_what
+        Executor.ncores
+        """
+        if (isinstance(workers, tuple)
+            and all(isinstance(i, (str, tuple)) for i in workers)):
+            workers = list(workers)
+        if workers is not None and not isinstance(workers, (list, set)):
+            workers = [workers]
+        return valmap(set, sync(self.loop, self.scheduler.processing,
+                                workers=workers))
+
     def nbytes(self, keys=None, summary=True):
         """ The bytes taken up by each key on the cluster
 
@@ -1538,12 +1612,18 @@ def redict_collection(c, dsk):
 
 
 def futures_of(o):
-    if isinstance(o, WrappedKey):
-        return [o]
-    if isinstance(o, (tuple, set, list)):
-        return [f for item in o for f in futures_of(item)]
-    if isinstance(o, dict):
-        return [f for v in o.values() for f in futures_of(v)]
-    if hasattr(o, 'dask'):
-        return futures_of(o.dask)
-    return []
+    """ Future objects in a collection """
+    stack = [o]
+    futures = set()
+    while stack:
+        x = stack.pop()
+        if type(x) in (tuple, set, list):
+            stack.extend(x)
+        if type(x) is dict:
+            stack.extend(x.values())
+        if type(x) is Future:
+            futures.add(x)
+        if hasattr(x, 'dask'):
+            stack.extend(x.dask.values())
+
+    return list(futures)
